#ifndef _SIF_H
#define _SIF_H

#include "siferrors.h"
#include "config.h"


typedef
struct /* sif */
{
  unsigned char* mem; /* memory block */
  cell cell_mask; /* & pointers with this before adding to "mem" */
  cell character_mask; /* & pointers with this before adding to "mem" */

  /* these stay the same: */
  cell* stack;
  cell* rstack;
  character* word;
  character* heap;
  character* dictionary;

  int cur_stack, cur_rstack, cur_heap, cur_dictionary; /* these all change as the program runs */
  cell code_position; /* where we are executing */
  int(*input)(void*, void*, int);
  void(*output)(void*, void*, int);
  void(*flush)(void*);

  int state; /* one of STATE_COMPILE or STATE_RUN */
  int heap_dict_size;
} sif;

void init_dictionary(sif* s);


#define DICTIONARY_ENTRY_SIZE (4 + 2 * CELL_SIZE) /* bytes */
#define STATE_INTERPRET 0
#define STATE_COMPILE 1
#define NOT_IMMEDIATE 0
#define IS_IMMEDIATE 1


/* this macro takes a sif* (s) and a cell (CELL) and returns the real memory address that CELL is pointing to */
#define CELL_TO_CELLPOINTER(s, CELL)   ((cell*)(s->mem + (CELL & (s->cell_mask))))
#define CELL_TO_CHARACTERPOINTER(s, CELL)   ((character*)(s->mem + (CELL & (s->character_mask))))


/* Stack macros */
#define GET_TOP(s) (s->cur_stack == -1 ? bye_message(s, "stack underflow\n") : s->stack[s->cur_stack])
#define SET_TOP(s, B) {if(s->cur_stack == -1) bye_message(s, "stack underflow\n"); s->stack[s->cur_stack] = (B);}

#define RGET_TOP(s) (s->cur_rstack == -1 ? bye_message(s, "rstack underflow\n") : s->rstack[s->cur_rstack])
#define RSET_TOP(s, B) {if(s->cur_rstack == -1) bye_message(s, "rstack underflow\n"); s->rstack[s->cur_rstack] = (B);}

#define PUSH(s, B) {s->cur_stack += 1; if(s->cur_stack == STACK_SIZE) bye_message(s, "stack overflow\n"); s->stack[s->cur_stack] = (B);}
#define RPUSH(s, B) {s->cur_rstack += 1; if(s->cur_rstack == STACK_SIZE) bye_message(s, "rstack overflow\n"); s->rstack[s->cur_rstack] = (B);}

#define POP(s) (s->cur_stack == -1 ? bye_message(s, "stack underflow\n") : s->stack[s->cur_stack--])
#define RPOP(s) (s->cur_rstack == -1 ? bye_message(s, "rstack underflow\n") : s->rstack[s->cur_rstack--])

#define ALLIGN(A) (((int)(A) + CELL_SIZE - 1) & (~(CELL_SIZE - 1)))

#define COMPILE(s, B) { \
   s->cur_heap += CELL_SIZE - 1;  /* allign */               \
   s->cur_heap &= s->cell_mask;                              \
   if(s->cur_heap + s->cur_dictionary >= s->heap_dict_size)  \
      bye_message(s, "Heap Overflow");                       \
   *((cell*)(s->mem + s->cur_heap)) = (B);                   \
   s->cur_heap += CELL_SIZE;                                 \
}

#define COMPILE_CHAR(s, B) {                                 \
   if(s->cur_heap + s->cur_dictionary >= s->heap_dict_size)  \
      bye_message(s, "Heap overflow");                       \
   *((character*)(s->mem + s->cur_heap)) = (B);              \
   s->cur_heap += 1;                                         \
}

#define HERE(s)  (s->cur_heap)

/* this prints message then exits forth */
int bye_message(sif* s, const char* message);

/* this just exits forth */
int bye(sif* s);

/* initialise sif structure */
int sif_init(sif* s, void* mem, int size);

/* make a 32 bit hash from a string */
const int hash_name(const char* name, int len);

#endif /* ifndef _SIF_H */
