
/* words.c
 *
 * This file contains all the forth words that are written in C
 *
 * The script dict.awk creates other source files according to this
 * one. If you do not follow the conventions below THIS WILL NOT WORK.
 *
 * 1) Functions starting with i_ are put in the forth dictionary as
 * ordinary words (that are called at runtime, and at compile time a
 * call to them is compiled in.)
 *
 * 2) Functions starting with c_ MUST apear directly after a function
 * with the same name except starting with i_. This c_ function will
 * be called only at compile time, and the preceeding i_ function will
 * be called at runtime.
 */

#include "sif.h"
#include "words.h"
#include "builtins.h"


void i_return(sif*s)
{
  s->code_position = RPOP(s);
}

/* i_return is used only by ; so I can tell when a word definition ends (for SEE) */
void i_exit(sif*s)
{
  s->code_position = RPOP(s);
}

void i_branch(sif*s)
{
  s->code_position += CELL_SIZE;
  s->code_position = *CELL_TO_CELLPOINTER(s, s->code_position); /* impl */
}

/* this function branches if it get's a FALSE value */
void i_questionbranch(sif*s) /* ?branch */
{
  s->code_position += CELL_SIZE;
  if(!POP(s))
    s->code_position = *CELL_TO_CELLPOINTER(s, s->code_position); /* impl */
}

void c_if(sif*s)
{
  COMPILE(s, I_QUESTIONBRANCH);
  i_here(s);
  COMPILE(s, I_RETURN); /* THEN should overwrite this */
}

void c_then(sif*s)
{
  i_here(s);
  i_cell(s);
  i_minus(s);
  i_swap(s);
  i_store(s);
}

void c_begin(sif*s)
{
  i_here(s);
  i_cell(s);
  i_minus(s);
}

void i_roneplus(sif*s) /* r1+ */
{
  cell c = RGET_TOP(s);

  RSET_TOP(s, c + 1);
}

void c_questiondo(sif*s) /* ?do */
{
  COMPILE(s, I__DO);
  COMPILE(s, I_BRANCH);
  i_here(s); /* actual branch goes one past here */
  COMPILE(s, I_RETURN);
}

void i__do(sif*s)
{
  cell c = POP(s);
  
  RPUSH(s, POP(s));
  RPUSH(s, c);
}

void c_do(sif*s)
{
  i_here(s); /* actual branch goes one past here */
  COMPILE(s, I__DO);
}

/* if the top two cells on the rstack are the same drop them and return true (true means don't branch). otherwise return false  */
void i__loop(sif*s)
{
  cell a = RPOP(s);
  cell b = RGET_TOP(s);

  if(s, a == b) /* this could be if(a >= b), but then you might not notice that your code is crap :) */
    {
      RPOP(s);
      PUSH(s, 1); /* counter and range are equal, remove values from returnstack and return true (means don't branch). */
    }
  else
    {
      RPUSH(s, a); /* put it back */
      PUSH(s, 0) /* return false (means branch to the begining of the loop again */
    }
}
      
void c_loop(sif*s)
{
  cell c = POP(s);

  if(*CELL_TO_CELLPOINTER(s, c) == I_RETURN) /* this is how we can tell the loop started with ?DO not DO */
    {
      *CELL_TO_CELLPOINTER(s, c) = HERE(s);
    }

  COMPILE(s, I_RONEPLUS);
  COMPILE(s, I__LOOP); /* do every time ... */
  COMPILE(s, I_QUESTIONBRANCH);
  COMPILE(s, c);
}


void c_until(sif*s)
{
  COMPILE(s, I_QUESTIONBRANCH);
  i_compilecomma(s);
}

/* push a pointer to the begining of the dictionary */
void i_dict(sif*s)
{
  PUSH(s, (cell)((int)(s->dictionary) - (int)(s->mem) - (int)(s->cur_dictionary))); /* impl */
}

/* push the number of words on the dictionary */
void i_numwords(sif*s) /* #words */
{
  PUSH(s, s->cur_dictionary / DICTIONARY_ENTRY_SIZE); /* impl */
}

void i_dots(sif*s) /* .s */
{
  int a = 0;
  while(a <= s->cur_stack)
    {
      PUSH(s, s->stack[a]);
      i_dot(s);
      ++a;
    }
}

void i_cell(sif*s)
{
  PUSH(s, CELL_SIZE);
}

void i_char(sif*s)
{
  char a = 0;

  if(!s->input(s, &a, 1))
    {
      bye_message(s, ""); /*End of input");*/
    }

  PUSH(s, (cell)a);
}

void i_dot(sif*s) /* . */
{
  cell a = POP(s);
  character* word = s->word + MAX_WORD_SIZE; /* impl using word buffer for string->number */
  character c;

  if(a == 0)
    {
      c = '0';
      s->output(s, &c, 1); 
    }
  else
    {
      while(a)
	{
	  c = a % BASE;
	  a /= BASE;
	  *(--word) = c + '0';
	}
      s->output(s, word, s->word + MAX_WORD_SIZE - word);
    }

  c = ' ';
  s->output(s, &c, 1);
}

void i_emit(sif*s)
{
  char a = (char)POP(s);

  s->output(s, &a, 1);
  s->flush(s);
}

void i_fetch(sif*s) /* @ */
{
  SET_TOP(s, *CELL_TO_CELLPOINTER(s, GET_TOP(s)));
}

void i_store(sif*s) /* ! */
{
  cell a = POP(s);
 
  *CELL_TO_CELLPOINTER(s, a) = POP(s);
}

void i_cfetch(sif*s) /* c@ */
{
  cell c = 0;

  SET_TOP(s, (cell)*CELL_TO_CHARACTERPOINTER(s, GET_TOP(s)));
}

void i_cstore(sif*s) /* c! */
{
  cell a = POP(s);
  cell c = POP(s);
 
  *CELL_TO_CHARACTERPOINTER(s, a) = (char)c;
}

void i_plus(sif*s) /* + */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) + a);
}

void i_oneplus(sif*s) /* 1+ */
{
  cell a = GET_TOP(s);

  SET_TOP(s, 1 + a);
}

void i_minus(sif*s) /* - */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) - a);
}

void i_times(sif*s) /* * */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) * a);
}

void i_lshift(sif*s)
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) << a);
}

void i_rshift(sif*s)
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) >> a);
}

void i_divide(sif*s) /* / */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) / a);
}

void i_mod(sif*s)
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) % a);
}

void i_equals(sif*s) /* = */
{
  cell c = POP(s);
  c = c == GET_TOP(s);
  SET_TOP(s, c);
}

void i_zeroequals(sif*s) /* 0= */
{
  SET_TOP(s, !GET_TOP(s));
}

void i_greater(sif*s) /* > */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) > a);
}

void i_less(sif*s) /* < */
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) < a);
}

void i_and(sif*s)
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) & a);
}

void i_or(sif*s)
{
  cell a = POP(s);

  SET_TOP(s, GET_TOP(s) | a);
}

void i_not(sif*s)
{
  cell a = GET_TOP(s);

  SET_TOP(s, ~a);
}

void i_rgreater(sif*s) /* r> */
{
  cell c = RPOP(s);

  PUSH(s, c);
}

void i_greaterr(sif*s) /* >r */
{
  cell c = POP(s);

  RPUSH(s, c);
}

void i_rfetch(sif*s) /* r@ */
{
  cell c = RGET_TOP(s);

  PUSH(s, c);
}

void i_i(sif*s)
{
  cell c = RGET_TOP(s);

  PUSH(s, c);
}

void i_here(sif*s)
{
  PUSH(s, HERE(s)); /* impl */
}

void i_comma(sif*s) /* , */
{
  COMPILE(s, POP(s));
}

void i_ccomma(sif*s) /* c, */
{
  cell c = POP(s);
  COMPILE_CHAR(s, (char)c);
}

void i_allign(sif*s)
{
  s->cur_heap = (s->cur_heap + CELL_SIZE - 1) & s->cell_mask;
}

void i_swap(sif*s)
{
  cell a, b;

  b = POP(s);
  a = GET_TOP(s);

  SET_TOP(s, b);
  PUSH(s, a);
}

void i_over(sif*s)
{
  cell a, b;

  b = POP(s);
  a = GET_TOP(s);

  PUSH(s, b);
  PUSH(s, a);
}

void i_rot(sif*s)
{
  cell a, b, c;

  c = POP(s);
  b = POP(s);
  a = GET_TOP(s);

  SET_TOP(s, b);
  PUSH(s, c);
  PUSH(s, a);
}

void i_drop(sif*s)
{
  POP(s);
}

void i_dup(sif*s)
{
  cell a = GET_TOP(s); 
  PUSH(s, a);
}

void i_depth(sif*s)
{
  PUSH(s, s->cur_stack); /* impl */
}

void i_pick(sif*s)
{
  cell a = GET_TOP(s);

  if(a > s->cur_stack)
    bye_message(s, "PICK out of range");

  SET_TOP(s, ((cell*)s->mem)[s->cur_stack - a]); /* impl */
}


/* ( char -- addr u ) NONSTANDARD */
void i_word(sif*s)
{
  cell a = GET_TOP(s);
  int i;
  character delim = (char)a;
  character  c = delim;
  character* word = s->word;

  if(delim == ' ')
    {
      
      /* eat trailing whitespace and check for EOF */
      while(c <= ' ')
	if(!s->input(s, &c, 1))
	  bye_message(s, ""); /*End of Input");*/
      
      for(i = 0; c > ' ' && i < MAX_WORD_SIZE-1; ++i)
	{
	  *word++ = c;
	  if(!s->input(s, &c, 1))
	    c = ' ';
	}
    }
  else
    {
      
      /* eat trailing whitespace and check for EOF */
      while(c == delim)
	if(!s->input(s, &c, 1))
	  bye_message(s, ""); /*End of Input");*/
      
      for(i = 0; c != delim && i < MAX_WORD_SIZE-1; ++i)
	{
	  *word++ = c;
	  if(!s->input(s, &c, 1))
	    c = delim;
	}
    }

  *word = 0; /* for "Word Not Found: " */

  SET_TOP(s, (int)s->word - (int)s->mem); /* impl */
  PUSH(s, word - s->word);
}

/* ( addr u -- xt immediate? true | addr u false ) NONSTANDARD */
/* at interpret time immediate? is always true */
void i_find(sif*s)
{
  int word = 0;
  cell len = POP(s);
  cell orig_len = len;
  cell cptr = GET_TOP(s);
  character* addr = CELL_TO_CHARACTERPOINTER(s, cptr);
  character* dictionary;
  cell* dict_words;

  if(len > MAX_WORD_SIZE)
    {
      bye_message(s, "FIND called with count too large");
    }


  if(((cptr + len) & s->character_mask) != ((cptr & s->character_mask) + len)) /* impl */
     bye_message(s, "FIND called with string that overflows memory block");

  /* make a 32 bit hash */
  word = hash_name(addr, len);

  
  dictionary = s->dictionary - s->cur_dictionary; /* impl ... */
  while(dictionary < s->dictionary)
    {
      if(*((int*)dictionary) == word)
	{
	  dict_words = (cell*)(dictionary+4); /* points to compiletime and runtime "pointers" */
	  SET_TOP(s, dict_words[0]); /* push xt */
	  
	  if(s->state == STATE_INTERPRET)
	    {
	      PUSH(s, 1); /* "should this run now?": yes */
	    }
	  else
	    {
	      PUSH(s, dict_words[1]); /* push is_immediate */
	    }
	  
	  PUSH(s, 1); /* true: word found */
	  return;
	}
      dictionary += DICTIONARY_ENTRY_SIZE;
    }

  PUSH(s, orig_len);
  PUSH(s, 0); /* word NOT found */
}

void i_bl(sif*s)
{
  PUSH(s, (cell)' ');
}

/* ( addr u -- int true | false ) NONSTANDARD */
void i_tonumber(sif*s) /* >number */
{
  cell a = 0;
  int len = POP(s);
  character *c = CELL_TO_CHARACTERPOINTER(s, GET_TOP(s));

  while(len--)
    {
      if(*c >= '0' && *c <= '9')
	a = a * BASE + (*c++ - '0');
      else
	{
	  SET_TOP(s, 0); /* return false if it's not a number */
	  return;
	}
    }

  SET_TOP(s, a);
  PUSH(s, 1);
}

/* this is used by the C interpret loop */
void grr_execute(sif*s)
{
  RPUSH(s, s->code_position);
  s->code_position = POP(s);
}

void i_compilecomma(sif*s) /* compile, */
{
  COMPILE(s, POP(s));
}

/* this is intended to be compiled by c_literal, not called at runtime */
void i__literal(sif*s)
{
  s->code_position += CELL_SIZE;
  PUSH(s, *CELL_TO_CELLPOINTER(s, s->code_position)); /* impl */
}

void c_literal(sif*s)
{
  COMPILE(s, I__LITERAL);
  COMPILE(s, POP(s));
}

void c_postpone(sif*s)
{
  i_bl(s);
  i_word(s);
  i_find(s);
  if(!POP(s))
    bye_message(s, "postpone: Word Not Found");
  
  i_drop(s);
  i_compilecomma(s);
}

/* ( addr u -- hash...    returns 32 bits, however many cells that is */
void i_hash(sif*s)
{
  cell* ptr;
  cell len = POP(s);
  cell cptr;
  int bits = 4;
  int hash;

  if(len > MAX_WORD_SIZE)
    {
      bye_message(s, "HASH called with count too large");
    }

  cptr = POP(s);
  if(((cptr + len) & s->character_mask) != ((cptr & s->character_mask) + len)) /* impl */
     bye_message(s, "HASH called with string that overflows memory block");

  hash = hash_name(CELL_TO_CHARACTERPOINTER(s, cptr), len);

  ptr = (cell*) &hash;
  while(bits)
    {
      PUSH(s, *ptr++);
      bits -= CELL_SIZE;
    }
}

/* ( addr u xt -- ) adds a dictionary entry */
void i_define(sif*s)
{
  i__define(s);
  s->cur_dictionary += DICTIONARY_ENTRY_SIZE; /* impl */
}
  
/* ( addr u xt -- ) sets up a dictionary entry */
void i__define(sif*s)
{
  cell len, cptr;
  int bits = 4;
  character* dictionary;

  if(s->cur_dictionary + s->cur_heap + DICTIONARY_ENTRY_SIZE > s->heap_dict_size)
    bye_message(s, "heap/dictionary overflow");

  dictionary = s->dictionary - s->cur_dictionary - DICTIONARY_ENTRY_SIZE; /* impl */


  ((cell*)(dictionary+4))[0] = POP(s); /* impl */

  /* hash name and stick it in the first 4 bytes of the dictionary entry */
  len = POP(s);
  cptr = POP(s);
  if(((cptr + len) & s->character_mask) != ((cptr & s->character_mask) + len)) /* impl */
     bye_message(s, "HASH called with string that overflows memory block");

  *(int*)(dictionary) = hash_name(CELL_TO_CHARACTERPOINTER(s, cptr), len);
}

void i_immediate(sif*s)
{
  character* dictionary = s->dictionary - s->cur_dictionary;
  ((cell*)(dictionary+4))[1] = IS_IMMEDIATE; /* impl */
}

void i_colon(sif*s) /* : */
{
  s->state = STATE_COMPILE;
  i_bl(s);
  i_word(s);
  i_here(s);
  i__define(s);
}

void c_semicolon(sif*s) /* ; */
{
  COMPILE(s, I_RETURN);
  s->state = STATE_INTERPRET;
  s->cur_dictionary += DICTIONARY_ENTRY_SIZE; /* impl */
}
    

int quit(sif*s)
{
  char b[100] = {'W', 'o', 'r', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'o', 'u', 'n', 'd', ':', ' '};
  int i;
  
  cell c;


  for(;;)
    {
      if(s->cur_rstack) /* if there is stuff to execute */
	{
	  if(s->code_position < NUM_BUILTINS) /* impl ... */
	    {
	      c = s->code_position;
	      s->code_position = RPOP(s);
	      builtins[c](s);
	      s->code_position += CELL_SIZE;
	    }
	  else
	    {
	      RPUSH(s, s->code_position);
	      s->code_position = (*CELL_TO_CELLPOINTER(s, s->code_position));
	    }
	}
      else /* rstack empty, so we interpret */
	{
	  i_bl(s);
	  i_word(s);
	  i_find(s);
	  if(POP(s)) /* if it's a word */
	    {
	      
	      if(POP(s)) /* if we should execute it :) */
		{
		  grr_execute(s);
		}
	      else
		{
		  i_compilecomma(s);
		}
	    }
	  else
	    {
	      i_tonumber(s);
	      if(POP(s))
		{
		  if(s->state == STATE_COMPILE)
		    c_literal(s);
		}
	      else
		{
		  for(i = 16; i < 99; ++i)
		    {
		      b[i] = s->word[i-16];
		    }
		  b[99] = 0;
		  bye_message(s, b);
		}
	    }
	}
    }
  return 77;
}

void i_create(sif*s)
{
  cell c;

  /* fix... check for dictionary overflow */

  i_bl(s);
  i_word(s);
  i_here(s);
  i_define(s);

  COMPILE(s, I__LITERAL);
  c = HERE(s);
  COMPILE(s, c + 3 * CELL_SIZE);
  COMPILE(s, I_RETURN);
  COMPILE(s, I_RETURN); /* this is here so does> can replace these two with a branch */
}

void c_does(sif*s) /* does> */
{
  cell c;
  COMPILE(s, I__LITERAL);
  c = HERE(s);
  COMPILE(s, c + 2 * CELL_SIZE);
  COMPILE(s, I__DOES);
  COMPILE(s, I_RETURN);
}

/* this is executed when you run a word that had does> in it. */
/* the address where the stuff you put after does> is compiled is passed on the top of the stack. */
/* this word modifies the last word defined (must be made by create) so that it branches to the aforementioned "stuff" after loading the literal instead of just returning. */
void i__does(sif*s)
{
  cell* c;

  c = (cell*)(s->dictionary - s->cur_dictionary + 4); /* pointer to curxt in dictionary */
  c = CELL_TO_CELLPOINTER(s, *c);
  c += 2;
  *c = I_BRANCH;
  ++c;
  *c = POP(s);
}
